#ifndef detectorTowerAngleAxis_H
#define detectorTowerAngleAxis_H
#include "detectorTowerController.h"
#include "turboPmacAxis.h"

class detectorTowerAngleAxis : public turboPmacAxis {
  public:
    /**
     * @brief Construct a new detectorTowerAngleAxis
     *
     * @param pController         Pointer to the associated controller
     * @param axisNo              Index of the axis
     * @param liftAxis            Pointer to the associated lift axis
     */
    detectorTowerAngleAxis(detectorTowerController *pController, int axisNo);

    /**
     * @brief Destroy the detectorTowerAngleAxis
     *
     */
    virtual ~detectorTowerAngleAxis();

    /**
     * @brief Readout of some values from the controller at IOC startup
     *
     * @return asynStatus
     */
    asynStatus init();

    /**
     * @brief Implementation of the `stop` function from asynMotorAxis
     *
     * @param acceleration        Acceleration ACCEL from the motor record. This
     * value is currently not used.
     * @return asynStatus
     */
    asynStatus stop(double acceleration);

    /**
     * @brief Poll all detector tower axes, if this axis is the one with the
     * smallest index.
     *
     * We do not use the doPoll framework from sinqMotor here on purpose, since
     * we want to e.g. reset the motorStatusProblem for all axes at once at the
     * beginning of the poll.
     *
     * @param moving
     * @return asynStatus
     */
    asynStatus poll(bool *moving);
    asynStatus doPoll(bool *moving);

    /**
     * @brief Set the target value for the detector angle and trigger a position
     * collection cycle
     *
     * @param position
     * @param relative
     * @param min_velocity
     * @param max_velocity
     * @param acceleration
     * @return asynStatus
     */
    asynStatus doMove(double position, int relative, double min_velocity,
                      double max_velocity, double acceleration);

    /**
     * @brief Start a movement to the target positions of this axis and the
     * attached offset axes.
     *
     * @return asynStatus
     */
    asynStatus startCombinedMove();

    /**
     * @brief This axis type has an absolute encoder by default
     *
     * @return asynStatus
     */
    asynStatus readEncoderType();

    /**
     * @brief If the input value is true, move the axis into working position,
     * otherwise into changer position.
     *
     * @return asynStatus
     */
    asynStatus toggleWorkingChangerState(bool toChangingPosition);

    /**
     * @brief Overwrite the `reset` function so it doesn't do fast polls
     * afterwards.
     *
     * The
     *
     * @return asynStatus
     */
    asynStatus reset() { return doReset(); }

    /**
     * @brief Implementation of the `doReset` method of sinqAxis
     *
     * @return asynStatus
     */
    asynStatus doReset();

    /**
     * @brief Move the axis to the position `newOrigin`  and recalibrate
     *
     * When calling this function, the angle axis moves to `newOrigin` and the
     * hardware sets this position as the new origin.
     *
     * @param origin
     * @return asynStatus
     */
    asynStatus adjustOrigin(double newOrigin);

    /**
     * @brief Read out the beam radius
     *
     * @return double
     */
    double beamRadius() { return beamRadius_; }

    /**
     * @brief Return a pointer to the associated angle axis
     *
     * @return detectorTowerAngleAxis*
     */
    detectorTowerLiftAxis *liftAxis() { return liftAxis_; }

    /**
     * @brief Return a pointer to the associated support axis
     *
     * @return detectorTowerAngleAxis*
     */
    detectorTowerSupportAxis *supportAxis() { return supportAxis_; }

    // If true, either this axis or one of the detectorTowerLiftAxis
    // attached to it received a movement command.
    bool receivedTarget_;

    // If set to true, the virtual axis is about to start a deferred movement
    // (but is currently still collecting movement commands from its component
    // axes)
    bool startingDeferredMovement_;

    /*
    The collector cycle waits until this time in seconds has passed before
    starting a deferred movement
    */
    double deferredMovementWait_;

    /**
     * @brief Variable holding the axis error for later use
     *
     */
    int error_;

    /**
     * @brief Return a pointer to the axis controller
     */
    virtual detectorTowerController *pController() override { return pC_; };

  protected:
    detectorTowerController *pC_;
    detectorTowerLiftAxis *liftAxis_;
    detectorTowerSupportAxis *supportAxis_;
    double beamRadius_;

  private:
    friend class detectorTowerLiftAxis;
    friend class detectorTowerSupportAxis;
};

#endif
