/********************************************
 *  detectorTowerController.h
 *
 * Detector tower controller driver based on the asynMotorController class
 *
 * Stefan Mathis, September 2024
 ********************************************/

#ifndef detectorTowerController_H
#define detectorTowerController_H
#include "turboPmacController.h"

// Forward declaration of the axis classes to resolve the cyclic dependency
// between the controller and the axis .h-file. See
// https://en.cppreference.com/w/cpp/language/class.
class detectorTowerAngleAxis;
class detectorTowerLiftAxis;
class detectorTowerSupportAxis;

class detectorTowerController : public turboPmacController {

  public:
    /**
     * @brief Construct a new detectorTowerController object
     *
     * This controller object can handle both "normal" TurboPMAC axes created
    with the turboPmacAxis constructor and the special detectorTowerAngleAxis.
     *
     * @param portName            See sinqController constructor
     * @param ipPortConfigName    See sinqController constructor
     * @param numAxes             See sinqController constructor
     * @param movingPollPeriod    See sinqController constructor
     * @param idlePollPeriod      See sinqController constructor
     * @param comTimeout          When trying to communicate with the device,
    the underlying asynOctetSyncIO interface waits for a response until this
    time (in seconds) has passed, then it declares a timeout.
     */
    detectorTowerController(const char *portName, const char *ipPortConfigName,
                            int numAxes, double movingPollPeriod,
                            double idlePollPeriod, double comTimeout);

    virtual ~detectorTowerController();

    /**
     * @brief Overloaded function of turboPmacController
     *
     * @param pasynUser
     * @param value
     * @return asynStatus
     */
    asynStatus readInt32(asynUser *pasynUser, epicsInt32 *value);

    /**
     * @brief Overloaded function of turboPmacController
     *
     * @param pasynUser           Specify the axis via the asynUser
     * @param value               New value
     * @return asynStatus
     */
    asynStatus writeInt32(asynUser *pasynUser, epicsInt32 value);

    /**
     * @brief Overloaded function of turboPmacController
     *
     * @param pasynUser           Specify the axis via the asynUser
     * @param value               New value
     * @return asynStatus
     */
    asynStatus writeFloat64(asynUser *pasynUser, epicsFloat64 value);

    /**
     * @brief Get the axis object
     *
     * @param pasynUser           Specify the axis via the asynUser
     * @return detectorTowerAngleAxis*        If no axis could be found, this is
     * a nullptr
     */
    detectorTowerAngleAxis *getDetectorTowerAngleAxis(asynUser *pasynUser);

    /**
     * @brief Get the axis object
     *
     * @param axisNo              Specify the axis via its index
     * @return detectorTowerAngleAxis*        If no axis could be found, this is
     * a nullptr
     */
    detectorTowerAngleAxis *getDetectorTowerAngleAxis(int axisNo);

    /**
     * @brief Get the axis object
     *
     * @param pasynUser           Specify the axis via the asynUser
     * @return detectorTowerLiftAxis*        If no axis could be found,
     * this is a nullptr
     */
    detectorTowerLiftAxis *getDetectorTowerLiftAxis(asynUser *pasynUser);

    /**
     * @brief Get the axis object
     *
     * @param axisNo              Specify the axis via its index
     * @return detectorTowerLiftAxis*        If no axis could be found,
     * this is a nullptr
     */
    detectorTowerLiftAxis *getDetectorTowerLiftAxis(int axisNo);

    /**
     * @brief Get the axis object
     *
     * @param pasynUser           Specify the axis via the asynUser
     * @return detectorTowerSupportAxis*        If no axis could be found,
     * this is a nullptr
     */
    detectorTowerSupportAxis *getDetectorTowerSupportAxis(asynUser *pasynUser);

    /**
     * @brief Get the axis object
     *
     * @param axisNo              Specify the axis via its index
     * @return detectorTowerSupportAxis*        If no axis could be found,
     * this is a nullptr
     */
    detectorTowerSupportAxis *getDetectorTowerSupportAxis(int axisNo);

    /**
     * @brief Common poll function for both lift and angle axes
     *
     * @param moving
     * @param angleAxis
     * @param liftAxis
     * @return asynStatus
     */
    asynStatus pollDetectorAxes(bool *moving, detectorTowerAngleAxis *angleAxis,
                                detectorTowerLiftAxis *liftAxis,
                                detectorTowerSupportAxis *supportAxis);

    // Accessors for additional PVs
    int positionStateRBV() { return positionStateRBV_; }
    int changeState() { return changeState_; }
    int changeStateRBV() { return changeStateRBV_; }
    int motorOrigin() { return motorOrigin_; }
    int motorAdjustOrigin() { return motorAdjustOrigin_; }
    int motorAdjustOriginHighLimitFromDriver() {
        return motorAdjustOriginHighLimitFromDriver_;
    }
    int motorAdjustOriginLowLimitFromDriver() {
        return motorAdjustOriginLowLimitFromDriver_;
    }

  private:
    // Indices of additional PVs
#define FIRST_detectorTower_PARAM positionStateRBV_
    int positionStateRBV_;
    int changeState_;
    int changeStateRBV_;
    int motorOrigin_;
    int motorAdjustOrigin_;
    int motorAdjustOriginHighLimitFromDriver_;
    int motorAdjustOriginLowLimitFromDriver_;
#define LAST_detectorTower_PARAM motorAdjustOriginLowLimitFromDriver_
};
#define NUM_detectorTower_DRIVER_PARAMS                                        \
    (&LAST_detectorTower_PARAM - &FIRST_detectorTower_PARAM + 1)

#endif /* detectorTowerController_H */
