#ifndef detectorTowerLiftAxis_H
#define detectorTowerLiftAxis_H
#include "detectorTowerController.h"
#include "turboPmacAxis.h"

class detectorTowerLiftAxis : public turboPmacAxis {
  public:
    /**
     * @brief Construct a new detectorTowerAngleAxis
     *
     * @param pController         Pointer to the associated controller
     * @param axisNo              Index of the axis
     */
    detectorTowerLiftAxis(detectorTowerController *pController, int axisNo,
                          detectorTowerAngleAxis *angleAxis);

    virtual ~detectorTowerLiftAxis();

    /**
     * @brief Readout of some values from the controller at IOC startup
     *
     * @return asynStatus
     */
    asynStatus init();

    /**
     * @brief Implementation of the `stop` function from asynMotorAxis
     *
     * @param acceleration        Acceleration ACCEL from the motor record. This
     * value is currently not used.
     * @return asynStatus
     */
    asynStatus stop(double acceleration);

    /**
     * @brief Poll all detector tower axes, if this axis is the one with the
     * smallest index.
     *
     * We do not use the doPoll framework from sinqMotor here on purpose, since
     * we want to e.g. reset the motorStatusProblem for all axes at once at the
     * beginning of the poll.
     *
     * @param moving
     * @return asynStatus
     */
    asynStatus poll(bool *moving);
    asynStatus doPoll(bool *moving);

    /**
     * @brief Set the target value for the detector lift and trigger a position
     * collection cycle
     *
     * @param position
     * @param relative
     * @param min_velocity
     * @param max_velocity
     * @param acceleration
     * @return asynStatus
     */
    asynStatus doMove(double position, int relative, double min_velocity,
                      double max_velocity, double acceleration);

    /**
     * @brief Calls the `reset` function of the associated angle axis.
     *
     * @param on
     * @return asynStatus
     */
    asynStatus reset();

    /**
     * @brief Move the axis to the position `newOrigin`  and recalibrate
     *
     * When calling this function, the lift axis moves to `newOrigin` and the
     * hardware sets this position as the new origin.
     *
     * @param origin
     * @return asynStatus
     */
    asynStatus adjustOrigin(double newOrigin);

    /**
     * @brief Move the support motor of the beam lift to the position
     * `newOrigin`  and recalibrate
     *
     * When calling this function, the lift axis support motor moves to
     * `newOrigin` and the hardware sets this position as the new origin.
     *
     * @param origin
     * @return asynStatus
     */
    asynStatus adjustSupportOrigin(double newOrigin);

    /**
     * @brief Return a pointer to the associated angle axis
     *
     * @return detectorTowerAngleAxis*
     */
    detectorTowerAngleAxis *angleAxis() { return angleAxis_; }

    /**
     * @brief This axis type has an absolute encoder by default
     *
     * @return asynStatus
     */
    asynStatus readEncoderType();

    /**
     * @brief Return a pointer to the axis controller
     */
    virtual detectorTowerController *pController() override { return pC_; };

  protected:
    detectorTowerController *pC_;
    detectorTowerAngleAxis *angleAxis_;

  private:
    friend class detectorTowerAngleAxis;
};

#endif
