#ifndef detectorTowerSupportAxis_H
#define detectorTowerSupportAxis_H
#include "detectorTowerController.h"
#include "turboPmacAxis.h"

/**
 * @brief Passive axis which is mostly controlled indirectly by the hardware
 *
 */
class detectorTowerSupportAxis : public turboPmacAxis {
  public:
    /**
     * @brief Construct a new detectorTowerSupportAxis
     *
     * @param pController         Pointer to the associated controller
     * @param axisNo              Index of the axis
     * @param liftAxis            Pointer to the associated lift axis
     */
    detectorTowerSupportAxis(detectorTowerController *pController, int axisNo,
                             detectorTowerAngleAxis *angleAxis);

    virtual ~detectorTowerSupportAxis();

    /**
     * @brief Readout of some values from the controller at IOC startup
     *
     * @return asynStatus
     */
    asynStatus init();

    /**
     * @brief Implementation of the `stop` function from asynMotorAxis
     *
     * @param acceleration        Acceleration ACCEL from the motor record. This
     * value is currently not used.
     * @return asynStatus
     */
    asynStatus stop(double acceleration);

    /**
     * @brief Poll all detector tower axes, if this axis is the one with the
     * smallest index.
     *
     * We do not use the doPoll framework from sinqMotor here on purpose, since
     * we want to e.g. reset the motorStatusProblem for all axes at once at the
     * beginning of the poll.
     *
     * @param moving
     * @return asynStatus
     */
    asynStatus poll(bool *moving);
    asynStatus doPoll(bool *moving);

    /**
     * @brief This axis cannot be moved directly
     *
     * @param position
     * @param relative
     * @param min_velocity
     * @param max_velocity
     * @param acceleration
     * @return asynStatus
     */
    asynStatus doMove(double position, int relative, double min_velocity,
                      double max_velocity, double acceleration) {
        return asynSuccess;
    }

    /**
     * @brief If the input value is true, move the axis into working position,
     * otherwise into changer position.
     *
     * @return asynStatus
     */
    asynStatus toggleWorkingChangerState(bool toChangingPosition);

    /**
     * @brief Calls the `reset` function of the associated lift axis.
     *
     * @param on
     * @return asynStatus
     */
    asynStatus reset();

    /**
     * @brief Move the axis to the position `newOrigin`  and recalibrate
     *
     * When calling this function, the angle axis moves to `newOrigin` and the
     * hardware sets this position as the new origin.
     *
     * @param origin
     * @return asynStatus
     */
    asynStatus adjustOrigin(double newOrigin);

    /**
     * @brief Return a pointer to the associated angle axis
     *
     * @return detectorTowerAngleAxis*
     */
    detectorTowerAngleAxis *angleAxis() { return angleAxis_; }

    /**
     * @brief This axis type has an absolute encoder by default
     *
     * @return asynStatus
     */
    asynStatus readEncoderType();

    /**
     * @brief Return a pointer to the axis controller
     */
    virtual detectorTowerController *pController() override { return pC_; };

  protected:
    detectorTowerController *pC_;
    detectorTowerAngleAxis *angleAxis_;

  private:
    friend class detectorTowerLiftAxis;
};

#endif
